(ns clojureDemo.VideoTransforms)
  
(import '(java.awt Rectangle Robot Toolkit) )
(import '(java.awt.image BufferedImage) )
(import '(java.awt Graphics2D Panel))
(import '(java.io File) )
(import '(javax.imageio ImageIO) )
(import '(javax.swing JFrame))
(import '(org.apache.commons.io FileUtils))
(import clojure.lang.LazySeq)
(import '(name.audet.samuel.javacv.jna highgui cv cxcore))
(import '(name.audet.samuel.javacv CanvasFrame))
(import '(name.audet.samuel.javacv.jna cxcore$IplImage))
(import '(name.audet.samuel.javacv.jna highgui$CvCapture$PointerByReference))
(import '(name.audet.samuel.javacv.jna highgui$CvVideoWriter$PointerByReference))
(import '(name.audet.samuel.javacv.jna cxcore$IplImage$PointerByReference))
(import '(name.audet.samuel.javacv.jna cxcore$IplImage))
(import '(name.audet.samuel.javacv JavaCvErrorCallback))

(.redirectError (JavaCvErrorCallback.));this enables the c errors to travel up to the JVM
                                       ;where they can be handled.


(use '[clojureDemo.VisionCore :only (video-seq cache video-data close-writer)])

 
(use 'clojure.contrib.repl-utils)

(def -inf Double/NEGATIVE_INFINITY)
(def inf  Double/POSITIVE_INFINITY)


(def lian (File. "/home/r/Desktop/source-videos/lian1.mpeg"))
(def look (File. "/home/r/Desktop/source-videos/dramatic_look.flv"))
(def getto(File. "/home/r/Desktop/source-videos/Ghetto.flv"))
(def human0(File. "/home/r/Desktop/source-videos/vsr1/human0.avi"))

(def base (File. "/home/r/Desktop/source-videos/"))

(def app0 (File. base "approach0v2.avi"))
(def app1 (File. base "approach1v3.avi"))
(def app2 (File. base "approach0v3.avi"))
(def app3 (File. base "approach2v2.avi"))
(def app4 (File. base "approach1v2.avi")) 
(def app5 (File. base "approach2v3.avi"))
  
(def bounce0 (File. base "bounce0v2.avi"))  
(def bounce1 (File. base "bounce1v3.avi")) 
(def bounce2 (File. base "bounce3v2.avi"))
(def bounce3 (File. base "bounce0v3.avi"))  
(def bounce4 (File. base "bounce2v2.avi"))
(def bounce5 (File. base "bounce1v2.avi"))  
(def bounce6 (File. base "bounce2v3.avi"))

(def collide0 (File. base "collide0v3.avi"))  
(def collide1 (File. base "collide2v3.avi")) 
(def collide2 (File. base "collide1v2.avi"))
(def collide3 (File. base "collide0v2.avi"))  
(def collide4 (File. base "collide1v3.avi")) 
  
(def give0 (File. base "give0v3.avi"))  
(def give1 (File. base "give2v3.avi")) 
(def give2 (File. base "give1v2.avi")) 
(def give3 (File. base "give0v2.avi"))        
(def give4 (File. base "give1v3.avi"))


(def target (File. "/home/r/Desktop/output-vision/"))
(def default(File. target "default.avi"))
(defn- makePanel [image] (proxy [Panel] [] (paint [g]  (.drawImage g image 0 0 nil))))

(defmulti  display "Creates a JFrame and displays a buffered image"  class)

(defmethod display 
  BufferedImage  [image] 
  (let [panel (makePanel image)
	frame (JFrame. "Oh Yeah!")]
    (.add frame panel) 
    (.pack frame) 
    (.setVisible frame true ) 
    (.setSize frame(.getWidth image) (.getHeight image))))

(defmethod display
  cxcore$IplImage [image]
  ( display (.getBufferedImage image)))
  
(defmethod display
  String [image]
  (display (highgui/cvLoadImage image highgui/CV_LOAD_IMAGE_COLOR)))

(defmethod display
  LazySeq [s]
  (display (first s)))


(defn video-writer
  "uses data about the video to make a writer"
  [data fileTarget]
  (highgui/cvCreateVideoWriter 
   (str fileTarget)
   
   ;(highgui/CV_FOURCC \P,\I,\M,\1) ;= MPEG-1 codec  (112913.386195 msecs) (104 MB)
   ;(highgui/CV_FOURCC \M,\J,\P,\G) ;= motion-jpeg codec (crashed)
   ;(highgui/CV_FOURCC \M,\P,\4,\2) ;= MPEG-4.2 codec (107184.186774 msecs) (82 MB)
   ;(highgui/CV_FOURCC \D,\I,\V,\3) ;= MPEG-4.3 codec (118308.933328 msecs)  (83 MB)
   ;;(highgui/CV_FOURCC \D,\I,\V,\X) ;= MPEG-4 codec    (99037.738131 msecs)  (85 MB)
   (highgui/CV_FOURCC \H,\D,\Y,\C)
   ;(highgui/CV_FOURCC \U,\2,\6,\3) ;= H263 codec     (101141.993551 msecs) (89 MB)
   ;(highgui/CV_FOURCC \I,\2,\6,\3) ;= H263I codec      (crashed)
   ;(highgui/CV_FOURCC \F,\L,\V,\1) ;= FLV1 codec     (104307.567802 msecs) (93 MB)
   ;(:codec data)                    ;= whatever the movie originally had.  (98278.694169 msecs) (1.9 GB)  

   (:fps data) (cxcore/cvSize (:width data) (:height data)) 
   1; 1 here means that we're writing in color.
    ; you cannot change it to 0 to write in 
    ; black and white. Everything just crashes instead.
    ; what a useful paramater.
   ))


(defn naturals []  (iterate inc 0))


(defn write-frame
  [writer frame]
  (do
    (highgui/cvWriteFrame writer frame)
    frame))

(defn number-seq
  [video-seq]
  (map #(vector %1 %2) (naturals) video-seq))

(defn save-seq 
  [writer video-seq]
  (map #(write-frame writer %) video-seq))
  
(defn create-runonce [function]
  (let [sentinel (Object.)
	result (atom sentinel)]
    (fn [& args]
      (locking sentinel
	(if (= @result sentinel)
	  (reset! result (function))
	  @result)))))

(defmacro oncer 
  [video-seq-gen]
  `((create-runonce #(~@video-seq-gen))))

(defmacro trans-save
"there's a small problem with trans-save --- it IS
truly transitive, but it does too much work....
sometimes it writes files twice.
this is functionally correct though."
  [target config video-seq]
  `(let [writer# (video-writer ~config ~target)]
     (do
       (dorun (save-seq writer# ~video-seq))
       (close-writer writer#)
       ~video-seq)))

(defn save-video
  [video target]
  (let [writer (video-writer (video-data video) target)]
    (do 
      (dorun (map #(write-frame writer %) (video-seq video)))
      (close-writer writer))))


(comment (Examples of things you can try that will actually work)

(def lazy-human (video-seq human0)) ;makes a lazy sequence of frames and returns instantly.
(def target1 (File. "some/path/out1.avi")) ;just creates a normal Java File object.
(def target2 (File. "some/other/path/out2.avi"))
(def human0-data (video-data human0)) ;creates a map containing the fps, width, and height of the video. 

(trans-save target human0-data (video-seq human0))
;saves a copy of human0 to disk.

(trans-save target2 human0-data (video-seq-filter (trans-save target1 human0-data (video-seq human0))))
;saves an unaltered copy of human0 to disk, filters the sequence of 
;Intel Processing Library images by video-seq-filter, and writes the 
;filtered result to disk.  video-seq-filter could discard every other frame,
;take the sequence by fives and do temporal blurring, or just turn every
;frame to black and white. 


(do (use :reload-all 'clojureDemo.VideoTransforms) (in-ns 'clojureDemo.VideoTransforms))

)

 

